#include "reliable_utils.h"

#define maxsobufsize       0x11000
int sndbufsize = maxsobufsize;
int rcvbufsize = maxsobufsize;


static int
_open_dgram (struct sockaddr_in *sin, u_int16_t port)
{
  size_t bufsize = maxsobufsize;
  socklen_t slen = sizeof (struct sockaddr_in);
  int s, n;

  if ((s = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    fprintf (stderr, "_open_dgram: could not allocate socket: %s\n", 
	     strerror (errno));
    return -1;
  }
  
  if (setsockopt (s, SOL_SOCKET, SO_SNDBUF, 
		  (char *) &bufsize, sizeof (bufsize)) < 0) {
    fprintf (stderr, "_open_dgram: setsockopt: SNDBUF: %s\n", strerror (errno));
    close (s);
    return -1;
  }

  if (setsockopt (s, SOL_SOCKET, SO_RCVBUF, 
		  (char *) &bufsize, sizeof (bufsize)) < 0) {
    fprintf (stderr, "_open_dgram: setsockopt: RCVBUF: %s\n", strerror (errno));
    close (s);
    return -1;
  }

  if ((n = fcntl (s, F_GETFL)) < 0
      || fcntl (s, F_SETFL, n | O_NONBLOCK) < 0) {
    fprintf (stderr, "_open_dgram: fcntl: NONBLOCK: %s\n", strerror (errno));
    close (s);
    return -1;
  }

  bzero (sin, slen);
  sin->sin_family      = AF_INET;
  sin->sin_port        = htons (port);
  sin->sin_addr.s_addr = htonl (INADDR_ANY);

  if (bind (s, (struct sockaddr *) sin, slen) < 0) {
    fprintf (stderr, "_open_dgram: bind: %s\n", strerror (errno));
    bzero (sin, sizeof (*sin));
    close (s);
    return -1;
  }

  if (getsockname(s, (struct sockaddr *) sin, &slen) < 0) {
    fprintf (stderr, "_open_dgram: getsockname: %s\n", strerror (errno));
    bzero (sin, sizeof (*sin));
    close (s);
    return -1;
  }    

  if (DEBUG_LVL & DEBUG_INIT) {
    fprintf (stderr, "_open_dgram: bound dgram: %s:%d fd [%d]\n",
	    inet_ntoa (sin->sin_addr), ntohs (sin->sin_port), s);
  }
  return s;
}


void
close_on_exec (int s)
{
  if (fcntl (s, F_SETFD, 1) < 0) {
    if (DEBUG_LVL & DEBUG_ERRORS)
      fprintf (stderr, "F_SETFD: %s\n", strerror (errno));
    exit (1);
  }
}

static int
_make_async (int s)
{
  int n;
  if ((n = fcntl (s, F_GETFL)) < 0
      || fcntl (s, F_SETFL, n | O_NONBLOCK) < 0)
    return -1;
  return 0;
}

void
make_async (int s)
{
  int n;
  int type;
  socklen_t sn;
  if (_make_async (s) < 0) {
    fprintf (stderr, "O_NONBLOCK: %s\n", strerror (errno));
    exit (1);
  }
  type = 0;
  sn = sizeof (type);
  if (getsockopt (s, SOL_SOCKET, SO_TYPE, (char *)&type, &sn) < 0)
    return;
#if defined (SO_RCVBUF) && defined (SO_SNDBUF)
  if (type == SOCK_STREAM)
    n = rcvbufsize;
  else
    n = maxsobufsize;
  if (setsockopt (s, SOL_SOCKET, SO_RCVBUF, (char *) &n, sizeof (n)) < 0)
    fprintf (stderr, "SO_RCVBUF: %s\n", strerror (errno));

  if (type == SOCK_STREAM)
    n = sndbufsize;
  else
    n = maxsobufsize;
  if (setsockopt (s, SOL_SOCKET, SO_SNDBUF, (char *) &n, sizeof (n)) < 0)
    fprintf (stderr, "SO_SNDBUF: %s\n", strerror (errno));
#endif /* SO_RCVBUF && SO_SNDBUF */
  /* Enable keepalives to avoid buffering write data for infinite time */
  n = 1;
  if (type == SOCK_STREAM
      && setsockopt (s, SOL_SOCKET, SO_KEEPALIVE,
		     (char *) &n, sizeof (n)) < 0)
    fprintf (stderr, "SO_KEEPALIVE: %s\n", strerror (errno));
}


int
open_dgram (struct sockaddr_in *sin, u_int16_t port)
{
  int s = _open_dgram (sin, port);
  if (s < 0) {
    fprintf (stderr, "failed to allocate dgram socket\n");
    return -1;
  }

  make_async (s);
  close_on_exec (s);
  return s;
}

/* From BSD kernel */

u_short
in_cksum (u_char *pkt, int len)
{
  register int nleft = len;
  register u_short *w = (u_short *) pkt;
  register int sum = 0;
  u_short answer = 0;
  
  /*
   * Our algorithm is simple, using a 32 bit accumulator (sum), we add
   * sequential 16 bit words to it, and at the end, fold back all the
   * carry bits from the top 16 bits into the lower 16 bits.
   */
  while (nleft > 1) {
    sum += *w++;
    nleft -= 2;
  }
  
  /* mop up an odd byte, if necessary */
  if (nleft == 1) {
    *(u_char *) (&answer) = *(u_char *) w ;
    sum += answer;
  }
  
  /* add back carry outs from top 16 bits to low 16 bits */
  sum = (sum >> 16) + (sum & 0xffff);     /* add hi 16 to low 16 */
  sum += (sum >> 16);                     /* add carry */
  answer = ~sum;                          /* truncate to 16 bits */
  return (answer);
}



